from inc_cfg import *

# Basic recwav start/stop test
# Test starts a call, caller plays audio, callee records it using 'call recwav'
test_param = TestParam(
    "Call recwav start/stop basic test",
    [
        InstanceParam("callee", "--null-audio --max-calls=4"),
        InstanceParam("caller", "--null-audio --max-calls=4")
    ]
)



import time
import sys
import os
import inc_const as const
import inc_util as util
from inc_cfg import *

# Load configuration
cfg_file = util.load_module_from_file("cfg_file", ARGS[1])

# Get test script name to determine which test to run
test_script_name = os.path.basename(ARGS[1])


##############################################################################
# Helper function: Basic playwav start/stop test
##############################################################################
def run_600_playwav_basic(t):
    callee = t.process[0]
    caller = t.process[1]

    if caller.use_telnet:
        caller.send("call new " + t.inst_params[0].uri)
    else:
        caller.send("m")
        caller.send(t.inst_params[0].uri)
    caller.expect(const.STATE_CALLING)

    time.sleep(0.2)
    callee.expect(const.EVENT_INCOMING_CALL)
    if callee.use_telnet:
        callee.send("call answer 200")
    else:
        callee.send("a")
        callee.send("200")

    caller.expect(const.STATE_CONFIRMED)
    callee.expect(const.STATE_CONFIRMED)

    caller.sync_stdout()
    callee.sync_stdout()

    # Start playing wav file on caller side to call 0
    if caller.use_telnet:
        caller.send("call playwav start 0 wavs/input.44.wav")
    caller.expect("Playback started")

    time.sleep(1)

    caller.sync_stdout()
    callee.sync_stdout()

    if caller.use_telnet:
        caller.send("call playwav stop 0")
    caller.expect("Playback stopped")

    caller.sync_stdout()
    callee.sync_stdout()

    if caller.use_telnet:
        caller.send("call hangup")
    else:
        caller.send("h")

    caller.expect(const.STATE_DISCONNECTED)
    callee.expect(const.STATE_DISCONNECTED)


##############################################################################
# Helper function: Playwav with multiple calls
##############################################################################
def run_601_playwav_multi(t):
    callee1 = t.process[0]
    callee2 = t.process[1]
    caller = t.process[2]

    # Caller makes first call to callee1
    if caller.use_telnet:
        caller.send("call new " + t.inst_params[0].uri)
    else:
        caller.send("m")
        caller.send(t.inst_params[0].uri)
    caller.expect(const.STATE_CALLING)

    time.sleep(0.2)
    callee1.expect(const.EVENT_INCOMING_CALL)
    if callee1.use_telnet:
        callee1.send("call answer 200")
    else:
        callee1.send("a")
        callee1.send("200")

    caller.expect(const.STATE_CONFIRMED)
    callee1.expect(const.STATE_CONFIRMED)

    caller.sync_stdout()
    callee1.sync_stdout()

    # Caller makes second call to callee2
    if caller.use_telnet:
        caller.send("call new " + t.inst_params[1].uri)
    else:
        caller.send("m")
        caller.send(t.inst_params[1].uri)
    caller.expect(const.STATE_CALLING)

    time.sleep(0.2)
    callee2.expect(const.EVENT_INCOMING_CALL)
    if callee2.use_telnet:
        callee2.send("call answer 200")
    else:
        callee2.send("a")
        callee2.send("200")

    caller.expect(const.STATE_CONFIRMED)
    callee2.expect(const.STATE_CONFIRMED)

    caller.sync_stdout()
    callee1.sync_stdout()
    callee2.sync_stdout()

    # Start playing wav file on first call (call-id 0)
    if caller.use_telnet:
        caller.send("call playwav start 0 wavs/input.44.wav")
    caller.expect("Playback started")

    time.sleep(1)

    caller.sync_stdout()

    if caller.use_telnet:
        caller.send("call playwav stop 0")
    caller.expect("Playback stopped")

    caller.sync_stdout()

    if caller.use_telnet:
        caller.send("call playwav start 1 wavs/input.44.wav")
    caller.expect("Playback started")

    time.sleep(1)

    caller.sync_stdout()

    if caller.use_telnet:
        caller.send("call playwav stop 1")
    caller.expect("Playback stopped")

    caller.sync_stdout()
    callee1.sync_stdout()
    callee2.sync_stdout()

    if caller.use_telnet:
        caller.send("call hangup")
    else:
        caller.send("h")

    caller.expect(const.STATE_DISCONNECTED)
    callee2.expect(const.STATE_DISCONNECTED)

    caller.sync_stdout()
    callee1.sync_stdout()

    if caller.use_telnet:
        caller.send("call hangup")
    else:
        caller.send("h")

    caller.expect(const.STATE_DISCONNECTED)
    callee1.expect(const.STATE_DISCONNECTED)


##############################################################################
# Helper function: Playwav with call-id -1
##############################################################################
def run_602_playwav_minus1(t):
    callee = t.process[0]
    caller = t.process[1]

    if caller.use_telnet:
        caller.send("call new " + t.inst_params[0].uri)
    else:
        caller.send("m")
        caller.send(t.inst_params[0].uri)
    caller.expect(const.STATE_CALLING)

    time.sleep(0.2)
    callee.expect(const.EVENT_INCOMING_CALL)
    if callee.use_telnet:
        callee.send("call answer 200")
    else:
        callee.send("a")
        callee.send("200")

    caller.expect(const.STATE_CONFIRMED)
    callee.expect(const.STATE_CONFIRMED)

    caller.sync_stdout()
    callee.sync_stdout()

    # Start playing wav file with call-id -1 (should use current call)
    if callee.use_telnet:
        callee.send("call playwav start -1 wavs/input.44.wav")
    callee.expect("Playback started")

    time.sleep(1)

    caller.sync_stdout()
    callee.sync_stdout()

    if callee.use_telnet:
        callee.send("call playwav stop -1")
    callee.expect("Playback stopped")

    caller.sync_stdout()
    callee.sync_stdout()

    if caller.use_telnet:
        caller.send("call hangup")
    else:
        caller.send("h")

    caller.expect(const.STATE_DISCONNECTED)
    callee.expect(const.STATE_DISCONNECTED)

    caller.sync_stdout()
    callee.sync_stdout()

    # Test queueing behavior: start playback before call is established
    # Start playback with call-id -1 when no call is active (should queue)
    if caller.use_telnet:
        caller.send("call playwav start -1 wavs/input.44.wav")
    caller.expect("No active call, playback queued for call setup")

    caller.sync_stdout()

    # Now make a new call - playback should start automatically
    if caller.use_telnet:
        caller.send("call new " + t.inst_params[0].uri)
    else:
        caller.send("m")
        caller.send(t.inst_params[0].uri)
    caller.expect(const.STATE_CALLING)

    time.sleep(0.2)
    callee.expect(const.EVENT_INCOMING_CALL)
    if callee.use_telnet:
        callee.send("call answer 200")
    else:
        callee.send("a")
        callee.send("200")

    caller.expect("Dynamic playback auto-started")
    caller.expect(const.STATE_CONFIRMED)
    callee.expect(const.STATE_CONFIRMED)

    caller.sync_stdout()
    callee.sync_stdout()

    time.sleep(1)

    if caller.use_telnet:
        caller.send("call playwav stop -1")
    caller.expect("Playback stopped")

    caller.sync_stdout()
    callee.sync_stdout()

    if caller.use_telnet:
        caller.send("call hangup")
    else:
        caller.send("h")

    caller.expect(const.STATE_DISCONNECTED)
    callee.expect(const.STATE_DISCONNECTED)


##############################################################################
# Helper function: Basic recwav start/stop test
##############################################################################
def run_603_recwav_basic(t):
    callee = t.process[0]
    caller = t.process[1]

    if caller.use_telnet:
        caller.send("call new " + t.inst_params[0].uri)
    else:
        caller.send("m")
        caller.send(t.inst_params[0].uri)
    caller.expect(const.STATE_CALLING)

    time.sleep(0.2)
    callee.expect(const.EVENT_INCOMING_CALL)
    if callee.use_telnet:
        callee.send("call answer 200")
    else:
        callee.send("a")
        callee.send("200")

    caller.expect(const.STATE_CONFIRMED)
    callee.expect(const.STATE_CONFIRMED)

    caller.sync_stdout()
    callee.sync_stdout()

    if caller.use_telnet:
        caller.send("call playwav start 0 wavs/input.44.wav")

    # Start recording on callee side
    if callee.use_telnet:
        callee.send("call recwav start 0 wavs/temp.44.wav")
    callee.expect("Recording started")

    time.sleep(2)

    caller.sync_stdout()
    callee.sync_stdout()

    if caller.use_telnet:
        caller.send("call playwav stop 0")

    if callee.use_telnet:
        callee.send("call recwav stop 0")
    callee.expect("Recording stopped")

    caller.sync_stdout()
    callee.sync_stdout()

    if caller.use_telnet:
        caller.send("call hangup")
    else:
        caller.send("h")

    caller.expect(const.STATE_DISCONNECTED)
    callee.expect(const.STATE_DISCONNECTED)


##############################################################################
# Helper function: Error handling test
##############################################################################
def run_604_error_handling(t):
    callee = t.process[0]
    caller = t.process[1]

    caller.trace("Test 1: Invalid call-id (no call exists)")
    if caller.use_telnet:
        caller.send("call playwav start 0 wavs/input.44.wav")

    # Should get an error message
    time.sleep(0.5)
    caller.expect("is not active")

    caller.sync_stdout()

    caller.trace("Test 2: Non-existent file")
    if caller.use_telnet:
        caller.send("call new " + t.inst_params[0].uri)
    else:
        caller.send("m")
        caller.send(t.inst_params[0].uri)
    caller.expect(const.STATE_CALLING)

    time.sleep(0.2)
    callee.expect(const.EVENT_INCOMING_CALL)
    if callee.use_telnet:
        callee.send("call answer 200")
    else:
        callee.send("a")
        callee.send("200")

    caller.expect(const.STATE_CONFIRMED)
    callee.expect(const.STATE_CONFIRMED)

    caller.sync_stdout()
    callee.sync_stdout()

    # Try to play non-existent file
    if caller.use_telnet:
        caller.send("call playwav start 0 /nonexistent/file.wav")

    time.sleep(0.5)
    caller.expect("Failed to create player")

    caller.sync_stdout()

    caller.trace("Test 3: Invalid call-id (out of range)")
    if caller.use_telnet:
        caller.send("call playwav start 99 wavs/input.44.wav")

    time.sleep(0.5)
    caller.expect("Invalid call_id")

    caller.sync_stdout()

    # Test 4: Successfully play a file (to verify system still works)
    caller.trace("Test 4: Valid playback")
    if caller.use_telnet:
        caller.send("call playwav start 0 wavs/input.44.wav")

    time.sleep(1)

    caller.sync_stdout()

    if caller.use_telnet:
        caller.send("call playwav stop 0")

    caller.sync_stdout()
    callee.sync_stdout()

    # Test 5: Try to stop playback when none is active
    caller.trace("Test 5: Stop when no playback active")
    if caller.use_telnet:
        caller.send("call playwav stop 0")

    time.sleep(0.5)
    caller.expect("No active playback")

    caller.sync_stdout()

    if caller.use_telnet:
        caller.send("call hangup")
    else:
        caller.send("h")

    caller.expect(const.STATE_DISCONNECTED)
    callee.expect(const.STATE_DISCONNECTED)


##############################################################################
# Main test function - dispatches to appropriate helper based on script name
##############################################################################
def test_func(t):
    if "600_playwav_basic" in test_script_name:
        run_600_playwav_basic(t)
    elif "601_playwav_multi_call" in test_script_name:
        run_601_playwav_multi(t)
    elif "602_playwav_callid_minus1" in test_script_name:
        run_602_playwav_minus1(t)
    elif "603_recwav_basic" in test_script_name:
        run_603_recwav_basic(t)
    elif "604_playwav_error_handling" in test_script_name:
        run_604_error_handling(t)
    else:
        raise TestError("Unknown test script: " + test_script_name)


# Post function for test 603
def post_func(t):
    if "603_recwav_basic" in test_script_name:
        # Check if the recorded file exists
        output_file = "wavs/temp.44.wav"
        if os.path.exists(output_file):
            file_size = os.path.getsize(output_file)
            if file_size > 0:
                t.process[0].trace("Recording successful, file size: " + str(file_size) + " bytes")
                os.remove(output_file)
            else:
                raise TestError("Recording file is empty")
        else:
            raise TestError("Recording file was not created")


# Here where it all comes together
test = cfg_file.test_param
test.test_func = test_func
test.post_func = post_func



import inc_sip as sip
import inc_sdp as sdp

sdp = \
"""
v=0
o=- 0 0 IN IP4 127.0.0.1
s=pjmedia
c=IN IP4 127.0.0.1
t=0 0
m=audio 4000 RTP/AVP 0 101
a=ice-ufrag:1234
a=ice-pwd:1234567890123456789012
a=rtpmap:0 PCMU/8000
a=sendrecv
a=rtpmap:101 telephone-event/8000
a=fmtp:101 0-15
a=candidate:XX 1 UDP 1234 127.0.0.1 4000 typ host
"""

args = "--null-audio --use-ice --auto-answer 200 --max-calls 1"
include = ["a=ice-ufrag"]		   	# must have ICE
exclude = ["a=candidate:[0-9a-zA-Z]+ 2 UDP", 	# must not answer with 2 components
	   "ice-mismatch"		     	# must not mismatch
	  ]

sendto_cfg = sip.SendtoCfg( "caller sends only one component", 
			    pjsua_args=args, sdp=sdp, resp_code=200, 
			    resp_inc=include, resp_exc=exclude)




from inc_cfg import *

# Test playwav with multiple calls to verify call-id parameter
# This test creates two calls and verifies that playwav can target specific calls by ID
test_param = TestParam(
    "Call playwav with multiple calls (call-id test)",
    [
        InstanceParam("callee1", "--null-audio --max-calls=4"),
        InstanceParam("callee2", "--null-audio --max-calls=4"),
        InstanceParam("caller", "--null-audio --max-calls=4")
    ]
)



#
# Test that "norefersub" is absent from the Supported header when pjsua is
# started with --no-supported-norefersub (sets pjsua_config.no_refer_sub = PJ_FALSE).
# SIPp verifies the Supported header in the INVITE does NOT contain
# "norefersub" and fails the test if it does.
#
import inc_const as const

PJSUA = ["--null-audio --max-calls=1 --no-tcp --no-supported-norefersub $SIPP_URI"]

PJSUA_EXPECTS = [
		 [0, const.STATE_CONFIRMED,     ""],
		 [0, const.STATE_DISCONNECTED,  ""]
		]



#
# Test that "norefersub" is present in the Supported header (default behavior).
# pjsua_config.no_refer_sub defaults to PJ_TRUE, so pjsua advertises
# "norefersub" per RFC 4488.  SIPp (UAC) sends OPTIONS to pjsua (UAS) and
# verifies the Supported header in the 200 OK response.
#
import inc_const as const

PJSUA = ["--null-audio --no-tcp"]

PJSUA_EXPECTS = []



#
# Test that "norefersub" is absent from the Supported header when pjsua is
# started with --no-supported-norefersub (sets pjsua_config.no_refer_sub = PJ_FALSE).
# SIPp (UAC) sends OPTIONS to pjsua (UAS) and verifies the Supported header
# in the 200 OK does NOT contain "norefersub", failing if it does.
#
import inc_const as const

PJSUA = ["--null-audio --no-tcp --no-supported-norefersub"]

PJSUA_EXPECTS = []



from inc_cfg import *

# Basic playwav start/stop test with single call
# Test starts a call, plays a WAV file using 'call playwav start', then stops it
test_param = TestParam(
    "Call playwav start/stop basic test",
    [
        InstanceParam("callee", "--null-audio --max-calls=4"),
        InstanceParam("caller", "--null-audio --max-calls=4")
    ]
)



#
# Test that "norefersub" is present in the Supported header (default behavior).
# pjsua_config.no_refer_sub defaults to PJ_TRUE, so pjsua advertises
# "norefersub" per RFC 4488.  SIPp verifies the Supported header in the INVITE.
#
import inc_const as const

PJSUA = ["--null-audio --max-calls=1 --no-tcp $SIPP_URI"]

PJSUA_EXPECTS = [
		 [0, const.STATE_CONFIRMED,     ""],
		 [0, const.STATE_DISCONNECTED,  ""]
		]



#
# Test driver for alt-pjsua-uac-custom-sdp.xml
#
# alt_pjsua acts as UAC. It places a call to SIPp (UAS) using a custom SDP
# with audio (PCMU) and video (H263-1998). The call should reach
# STATE_CONFIRMED.
#
# Run with alt_pjsua:
#   cd tests/pjsua && python3 run.py \
#       --exe ../../pjsip-apps/src/3rdparty_media_sample/alt_pjsua \
#       mod_sipp.py scripts-sipp/alt-pjsua-uac-custom-sdp.xml
#
import inc_const as const

# pjsua instance: UAC. No URI in args so the call is initiated via CLI
# after cli_main() has set up the telnet log redirect.
# --rtp-port 4000 pins the audio RTP port; video uses 4002 (next even port).
# PJMEDIA_HAS_VIDEO=1 in config_site.h makes pjsua_call_setting_default set
# vid_cnt=1, so med_prov_cnt=2 (audio+video), matching the custom SDP.
PJSUA = [
    "--null-audio --max-calls=1 --no-tcp --rtp-port 4000 "
    "--custom-sdp \"v=0\\r\\no=- 1234567890 1234567890 IN IP4 127.0.0.1"
    "\\r\\ns=-\\r\\nc=IN IP4 127.0.0.1\\r\\nt=0 0"
    "\\r\\nm=audio 4000 RTP/AVP 0\\r\\na=rtpmap:0 PCMU/8000"
    "\\r\\nm=video 4002 RTP/AVP 96\\r\\na=rtpmap:96 H263-1998/90000\""
]

PJSUA_CLI_EXPECTS = [
    [0, "", "call new $SIPP_URI"],
    # Verify the SDP SIPp sends in its 200 OK is received by alt_pjsua.
    # Codec check only (no IP/port) for environment independence.
    [0, r"v=0", ""],
    [0, r"s=-", ""],
    [0, r"t=0 0", ""],
    [0, r"a=rtpmap:0 PCMU/8000", ""],
    [0, r"a=rtpmap:96 H263-1998/90000", ""],
    [0, const.STATE_CONFIRMED, "call hangup"],
]



#!/usr/bin/env python3
#
# importsym2.py: Import C symbol decls (structs, enums, typedefs) and write
#                them to another file. Python 3 rewrite of importsym.py
#                with no external dependencies (no pycparser needed).
#
# Copyright (C)2013-2026 Teluu Inc. (http://www.teluu.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
import subprocess
import shutil
import sys
import os
import re
import tempfile

PJ_ROOT_PATH = "../../../"

INCLUDE_DIRS = [
    PJ_ROOT_PATH + "pjlib/include",
    PJ_ROOT_PATH + "pjlib-util/include",
    PJ_ROOT_PATH + "pjnath/include",
    PJ_ROOT_PATH + "pjmedia/include",
    PJ_ROOT_PATH + "pjsip/include",
]


def find_cpp():
    """Find the C preprocessor."""
    cpp = shutil.which("cpp")
    if cpp:
        return cpp
    # On some systems, try gcc -E
    gcc = shutil.which("gcc")
    if gcc:
        return gcc
    cc = shutil.which("cc")
    if cc:
        return cc
    return None


def preprocess(header_files):
    """Run the C preprocessor on the given header files and return the
    preprocessed output as a string.
    """
    cpp = find_cpp()
    if not cpp:
        print("Error: need cpp, gcc, or cc in PATH")
        sys.exit(1)

    # Build a temporary C file that includes all the headers
    content = "#define PJ_AUTOCONF 1\n"
    content += "#define __attribute__(x)\n"
    content += "#define PJ_DEPRECATED(x)\n"
    content += "#define PJ_ATTR_DEPRECATED\n"
    for hf in header_files:
        content += '#include <%s>\n' % hf

    # Write to temp file
    fd, tmpfile = tempfile.mkstemp(suffix='.h')
    try:
        os.write(fd, content.encode())
        os.close(fd)

        # Build cpp command
        cmd = [cpp]
        if 'gcc' in os.path.basename(cpp) or 'cc' == os.path.basename(cpp):
            cmd.append('-E')
        for d in INCLUDE_DIRS:
            cmd.extend(['-I', d])
        cmd.append(tmpfile)

        result = subprocess.run(cmd, capture_output=True, text=True)
        if result.returncode != 0:
            print("Preprocessor error:")
            print(result.stderr)
            sys.exit(1)
        return result.stdout
    finally:
        os.unlink(tmpfile)


def find_matching_brace(source, start):
    """Find the position of the closing '}' that matches the '{' at start,
    handling nested braces. Returns the index of the closing '}'.
    """
    depth = 0
    i = start
    while i < len(source):
        if source[i] == '{':
            depth += 1
        elif source[i] == '}':
            depth -= 1
            if depth == 0:
                return i
        i += 1
    return -1


def extract_declarations(preprocessed, names):
    """Extract enum, struct, and typedef declarations for the given names
    from preprocessed C source.

    Returns a dict mapping name -> declaration string.
    """
    # Remove line directives (# 123 "file.h" ...)
    lines = []
    for line in preprocessed.split('\n'):
        if line.startswith('#'):
            continue
        lines.append(line)
    source = '\n'.join(lines)

    # Collapse multiple blank lines
    source = re.sub(r'\n{3,}', '\n\n', source)

    results = {}
    name_set = set(names)

    # Find all typedef/enum/struct declarations using brace matching
    # Pattern: typedef (enum|struct) NAME { ... } NAME ;
    for m in re.finditer(
            r'typedef\s+(enum|struct)\s+(\w+)\s*\{', source):
        kind = m.group(1)
        tag_name = m.group(2)
        brace_start = m.end() - 1  # position of '{'
        brace_end = find_matching_brace(source, brace_start)
        if brace_end < 0:
            continue
        # Find the alias name and semicolon after '}'
        rest = source[brace_end+1:brace_end+100]
        alias_m = re.match(r'\s*(\w+)\s*;', rest)
        if not alias_m:
            continue
        alias = alias_m.group(1)
        full_decl = source[m.start():brace_end+1+alias_m.end()]
        if tag_name in name_set or alias in name_set:
            key = tag_name if tag_name in name_set else alias
            if key not in results:
                results[key] = clean_decl(full_decl)

    # Pattern: enum NAME { ... };  (no typedef)
    for m in re.finditer(r'(?<!\w)enum\s+(\w+)\s*\{', source):
        enum_name = m.group(1)
        if enum_name not in name_set or enum_name in results:
            continue
        # Check this isn't preceded by 'typedef'
        prefix = source[max(0, m.start()-20):m.start()]
        if 'typedef' in prefix:
            continue
        brace_start = m.end() - 1
        brace_end = find_matching_brace(source, brace_start)
        if brace_end < 0:
            continue
        rest = source[brace_end+1:brace_end+20]
        semi_m = re.match(r'\s*;', rest)
        if not semi_m:
            continue
        full_decl = source[m.start():brace_end+1+semi_m.end()]
        results[enum_name] = clean_decl(full_decl)

    # Pattern: typedef <type> NAME;  (simple typedef)
    for m in re.finditer(r'typedef\s+(\w[\w\s]*?)\s+(\w+)\s*;', source):
        alias = m.group(2)
        if alias in name_set and alias not in results:
            results[alias] = clean_decl(m.group(0))

    return results


def clean_decl(decl):
    """Clean up a declaration: normalize whitespace, fix indentation,
    remove blank lines inside braces.
    """
    lines = decl.split('\n')
    cleaned = []
    for line in lines:
        line = line.rstrip()
        # Skip blank lines inside the declaration
        if line == '' and cleaned and cleaned[-1] != '':
            # Keep blank lines only between declarations, not inside braces
            in_braces = any('{' in l for l in cleaned) and \
                        not any('}' in l for l in cleaned)
            if in_braces:
                continue
        cleaned.append(line)

    # Remove any trailing blank lines before closing
    while len(cleaned) > 1 and cleaned[-2] == '':
        cleaned.pop(-2)

    result = '\n'.join(cleaned).strip()
    return result


def reformat_enum(decl):
    """Reformat an enum declaration for clean output."""
    # Already clean from cpp output, just ensure consistent style
    return decl


def process(listfile, outfile):
    """Main processing: read listfile, preprocess headers, extract symbols,
    write output.
    """
    # Read listfile
    with open(listfile) as f:
        lines = f.readlines()

    # Parse list file
    header_files = []
    names = []  # ordered list of names to extract
    for line in lines:
        spec = line.split()
        if len(spec) < 2:
            continue
        header_files.append(spec[0])
        names.extend(spec[1:])

    print('Preprocessing %d files for %d symbols..' %
          (len(header_files), len(names)))

    # Preprocess
    preprocessed = preprocess(header_files)

    # Extract declarations
    print('Extracting declarations..')
    decls = extract_declarations(preprocessed, names)

    # Write output in the order specified in listfile
    print('Writing declarations..')
    with open(outfile, 'w') as f:
        f.write("// This file is autogenerated by importsym script, "
                "do not modify!\n\n")
        for name in names:
            if name not in decls:
                print("  warning: declaration for '%s' is not found **" % name)
            else:
                print("  writing '%s'.." % name)
                decl = decls[name]
                # Avoid double semicolon
                if decl.endswith(';'):
                    f.write(decl + '\n\n')
                else:
                    f.write(decl + ';\n\n')

    print("Done. Output written to '%s'" % outfile)


if __name__ == "__main__":
    listfile = "symbols.lst"
    outfile = "symbols.i"

    if len(sys.argv) >= 3:
        listfile = sys.argv[1]
        outfile = sys.argv[2]

    print("Importing symbols: '%s' --> '%s'" % (listfile, outfile))
    process(listfile, outfile)



import inc_sip as sip
import inc_sdp as sdp

# When SRTP is enabled in pjsua, it should prefer to use
# RTP/SAVP media line if there are multiple m=audio lines
sdp = \
"""
v=0
o=- 0 0 IN IP4 127.0.0.1
s=-
c=IN IP4 127.0.0.1
t=0 0
m=audio 4000 RTP/AVP 0
a=rtpmap:0 pcmu/8000
m=audio 4000 RTP/SAVP 0
a=crypto:1 aes_cm_128_hmac_sha1_80 inline:WnD7c1ksDGs+dIefCEo8omPg4uO8DYIinNGL5yxQ
"""

pjsua_args = "--null-audio --auto-answer 200 --use-srtp 1 --srtp-secure 0"
extra_headers = ""
include = ["Content-Type: application/sdp",	# response must include SDP
	   "m=audio 0 RTP/AVP[\\s\\S]+m=audio [1-9]+[0-9]* RTP/SAVP[\\s\\S]+a=crypto"
	   ]
exclude = []

sendto_cfg = sip.SendtoCfg("Prefer RTP/SAVP", pjsua_args, sdp, 200,
			   extra_headers=extra_headers,
			   resp_inc=include, resp_exc=exclude) 





# Quality test of media calls.
# - UA1 calls UA2
# - UA1 plays a file until finished to be streamed to UA2
# - UA2 records from stream
# - Apply PESQ to played file (reference) and recorded file (degraded)
#
# File should be:
# - naming: xxxxxx.CLOCK_RATE.wav, e.g: test1.8.wav
# - clock-rate of those files can only be 8khz or 16khz

import time
import os
import sys
import re
import subprocess
import wave
import shutil
import inc_const as const
import inc_util as util
import numpy as np

from inc_cfg import *
from pesq import pesq

# Load configuration
cfg_file = util.load_module_from_file("cfg_file", ARGS[1])

# PESQ configs
PESQ = "tools/pesq"				# PESQ executable path
VISQOL = "tools/visqol"         # Visqol executable path
PESQ_DEFAULT_THRESHOLD = 3.4    # Default minimum acceptable PESQ MOS value
USE_PYTHON_PESQ = 1             # Use python's pesq
ENABLE_VISQOL = 1               # Enable Visqol
VISQOL_DEFAULT_THRESHOLD = 4    # Default minimum acceptable Visqol MOS value
VISQOL_MODEL_FILE = "tools/model/mod.tflite" # Visqol model file

# PESQ params
pesq_sample_rate_opt = ""		# Sample rate option for PESQ
input_filename    = ""			# Input/Reference filename
output_filename = ""			# Output/Degraded filename


# Test body function
def test_func(t):
    global pesq_sample_rate_opt
    global input_filename
    global output_filename
    global ENABLE_VISQOL

    ua1 = t.process[0]
    ua2 = t.process[1]

    # Get input file name
    input_filename = re.compile(const.MEDIA_PLAY_FILE).search(ua1.inst_param.arg).group(1)

    # Get output file name
    output_filename = re.compile(const.MEDIA_REC_FILE).search(ua2.inst_param.arg).group(1)

    # Get WAV input length, in seconds
    fin = wave.open(input_filename, "r")
    if fin == None:
        raise TestError("Failed opening input WAV file")
    inwavlen = fin.getnframes() * 1.0 / fin.getframerate()
    inwavlen += 0.2
    fin.close()
    print("WAV input len = " + str(inwavlen) + "s")

    # Get clock rate of the output
    mo_clock_rate = re.compile("\.(\d+)\.wav").search(output_filename)
    if (mo_clock_rate is None):
        raise TestError("Cannot compare input & output, incorrect output filename format")
    clock_rate = mo_clock_rate.group(1)
    # Get channel count of the output
    channel_count = 1
    if re.search("--stereo", ua2.inst_param.arg) is not None:
        channel_count = 2
    # Get matched input file from output file
    # (PESQ evaluates only files whose same clock rate & channel count)
    if channel_count == 2:
        if re.search("\.\d+\.\d+\.wav", input_filename) != None:
            input_filename = re.sub("\.\d+\.\d+\.wav", "." + str(channel_count) + "."+clock_rate+".wav", input_filename)
        else:
            input_filename = re.sub("\.\d+\.wav", "." + str(channel_count) + "."+clock_rate+".wav", input_filename)

    if (clock_rate != "8") and (clock_rate != "16"):
        raise TestError("PESQ only works on clock rate 8kHz or 16kHz, clock rate used = "+clock_rate+ "kHz")

    if (ENABLE_VISQOL == 1) and (clock_rate != "16"):
        print("VISQOL only works on clock rate 16kHz, clock rate used = "+ clock_rate +"kHz, disabling VISQOL")
        ENABLE_VISQOL = 0

    # Get conference clock rate of UA2 for PESQ sample rate option
    pesq_sample_rate_opt = "+" + clock_rate + "000"

    # UA1 making call
    if ua1.use_telnet:
        ua1.send("call new " + t.inst_params[1].uri)
    else:
        ua1.send("m")
        ua1.send(t.inst_params[1].uri)
    ua1.expect(const.STATE_CALLING)

    # UA2 wait until call established
    ua2.expect(const.STATE_CONFIRMED)

    ua1.sync_stdout()
    ua2.sync_stdout()
    time.sleep(2)

    # Disconnect mic -> rec file, to avoid echo recorded when using sound device
    # Disconnect stream -> spk, make it silent
    # Connect stream -> rec file, start recording
    ua2.send("cd 0 1")
    ua2.send("cd 4 0")
    ua2.send("cc 4 1")

    # Disconnect mic -> stream, make stream purely sending from file
    # Disconnect stream -> spk, make it silent
    # Connect file -> stream, start sending
    ua1.send("cd 0 4")
    ua1.send("cd 4 0")
    ua1.send("cc 1 4")

    time.sleep(inwavlen)

    # Disconnect files from bridge
    ua2.send("cd 4 1")
    ua2.expect(const.MEDIA_DISCONN_PORT_SUCCESS)
    ua1.send("cd 1 4")
    ua1.expect(const.MEDIA_DISCONN_PORT_SUCCESS)

def run_visqol(input_filename, output_filename):
    visqol_bin = os.path.abspath(VISQOL)
    model_file = os.path.abspath(VISQOL_MODEL_FILE)

    cmd = [
        visqol_bin,
        "--similarity_to_quality_model", model_file,
        "--use_speech_mode",
        "--reference_file", input_filename,
        "--degraded_file", output_filename
    ]

    try:
        process = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True
        )
        stdout, stderr = process.communicate()

        if process.returncode != 0:
            raise Exception(f"ViSQOL failed with exit code {process.returncode}: {stderr}")

        match = re.search(r"MOS-LQO:\s+([\d\.]+)", stdout)

        if not match:
            print(f"DEBUG OUTPUT: {stdout}")
            raise Exception("Failed to parse MOS-LQO from ViSQOL output")

        return float(match.group(1))

    except Exception as e:
        print("Visqol Error: " + str(e))
        raise

def read_wav_to_array(filename):
    with wave.open(filename, 'rb') as wf:
        rate = wf.getframerate()
        if rate != 8000 and rate != 16000:
            raise TestError("PESQ only works on clock rate 8kHz or 16kHz, clock rate used = "+str(rate)+ "kHz")

        # Get raw bytes and convert to 16-bit integers
        frames = wf.readframes(wf.getnframes())
        audio_array = np.frombuffer(frames, dtype=np.int16)

        return rate, audio_array


# Post body function
def post_func(t):
    global pesq_sample_rate_opt
    global input_filename
    global output_filename
    pesq_test_res = 0
    visqol_test_res = 0

    endpt = t.process[0]

    # Get threshold
    if (cfg_file.pesq_threshold is not None and cfg_file.pesq_threshold > -0.5 ):
        pesq_threshold = cfg_file.pesq_threshold
    else:
        pesq_threshold = PESQ_DEFAULT_THRESHOLD

    if (USE_PYTHON_PESQ == 1):
        # Evaluate the PESQ MOS value
        in_rate, in_data = read_wav_to_array(input_filename)
        out_rate, out_data = read_wav_to_array(output_filename)
        if (in_rate != out_rate):
            raise TestError("Different rate from input file = "+ in_rate + " and output file = " + out_rate)

        if in_rate == 8000:
            mode='nb'
        else:
            mode='wb'

        pesq_res = pesq(in_rate, in_data, out_data, mode)
    else:
        # Execute PESQ
        fullcmd = os.path.normpath(PESQ) + " " + pesq_sample_rate_opt + " " + input_filename + " " + output_filename
        endpt.trace("Popen " + fullcmd)
        pesq_proc = subprocess.Popen(fullcmd, shell=True, stdout=subprocess.PIPE, universal_newlines=True)
        pesq_out  = pesq_proc.communicate()

        # Parse output
        mo_pesq_out = re.compile("Prediction[^=]+=\s+([\-\d\.]+)\s*").search(pesq_out[0])
        if (mo_pesq_out is None):
            raise TestError("Failed to fetch PESQ result")

        pesq_res = mo_pesq_out.group(1)

    if (float(pesq_res) >= pesq_threshold):
        endpt.trace("Success, PESQ result = " + str(pesq_res) + " (target=" + str(pesq_threshold) + ").")
        pesq_test_res = 1
    else:
        endpt.trace("Failed, PESQ result = " + str(pesq_res) + " (target=" + str(pesq_threshold) + ").")

    if (ENABLE_VISQOL == 1):
        visqol_threshold = getattr(cfg_file, 'visqol_threshold', None)
        if (visqol_threshold is None) or (visqol_threshold <= 0 ):
            visqol_threshold = VISQOL_DEFAULT_THRESHOLD

        visqol_res = run_visqol(input_filename, output_filename)

        if (float(visqol_res) >= visqol_threshold):
            endpt.trace("Success, VISQOL result = " + str(visqol_res) + " (target=" + str(visqol_threshold) + ").")
            visqol_test_res = 1
        else:
            endpt.trace("Failed, VISQOL result = " + str(visqol_res) + " (target=" + str(visqol_threshold) + ").")

    if (pesq_test_res == 0 and visqol_test_res == 0):
        # Save the wav file
        wavoutname = ARGS[1]
        wavoutname = re.sub("[\\\/]", "_", wavoutname)
        wavoutname = re.sub("\.py$", ".wav", wavoutname)
        wavoutname = "logs/" + wavoutname
        try:
            shutil.copyfile(output_filename, wavoutname)
            print("Output WAV is copied to " + wavoutname)
        except:
            print("Couldn't copy output WAV, please check if 'logs' directory exists.")

        err_str = "WAV seems to be degraded badly, PESQ = "+ str(pesq_res) + " (target=" + str(pesq_threshold) + ")."
        if (ENABLE_VISQOL == 1):
            err_str += ", VISQOL = " + str(visqol_res) + " (target=" + str(visqol_threshold) + ")."

        raise TestError(err_str)


# Here where it all comes together
test = cfg_file.test_param
test.test_func = test_func
test.post_func = post_func



from inc_cfg import *

# Test error handling for invalid call-id and missing files
# Tests various error conditions:
# - Invalid call-id (no call exists)
# - Non-existent file
# - Invalid call-id (out of range)
# - Stop when no playback active
test_param = TestParam(
    "Call playwav error handling test",
    [
        InstanceParam("callee", "--null-audio --max-calls=4"),
        InstanceParam("caller", "--null-audio --max-calls=4")
    ]
)



from inc_cfg import *

# Test playwav with call-id -1 (current call or queue)
# Tests: 1) Using current call when call is active
#        2) Queueing for next call when no call is active
test_param = TestParam(
    "Call playwav with call-id -1 (current/queue)",
    [
        InstanceParam("callee", "--null-audio --max-calls=4"),
        InstanceParam("caller", "--null-audio --max-calls=4")
    ]
)



#
# Test driver for alt-pjsua-uas-custom-sdp.xml
#
# alt_pjsua (or pjsua) acts as UAS with --auto-answer=200. SIPp sends an
# INVITE with a PCMU offer. pjsua answers with its custom SDP. The call
# should reach STATE_CONFIRMED before SIPp terminates it with BYE.
#
# Run with standard pjsua:
#   cd tests/pjsua && python3 run.py mod_sipp.py \
#       scripts-sipp/alt-pjsua-uas-custom-sdp.xml
#
# Run with alt_pjsua:
#   cd tests/pjsua && python3 run.py \
#       --exe ../../pjsip-apps/src/3rdparty_media_sample/alt_pjsua \
#       mod_sipp.py scripts-sipp/alt-pjsua-uas-custom-sdp.xml
#
import inc_const as const

# pjsua instance: UAS, auto-answers with custom SDP.
# --rtp-port 4002 pins the audio RTP port; video uses 4004 (next even port).
PJSUA = [
    "--null-audio --max-calls=1 --no-tcp --rtp-port 4002 "
    "--auto-answer=200 "
    "--custom-sdp \"v=0\\r\\no=- 1234567890 1234567890 IN IP4 127.0.0.1"
    "\\r\\ns=-\\r\\nc=IN IP4 127.0.0.1\\r\\nt=0 0"
    "\\r\\nm=audio 4002 RTP/AVP 0\\r\\na=rtpmap:0 PCMU/8000"
    "\\r\\nm=video 4004 RTP/AVP 96\\r\\na=rtpmap:96 H263-1998/90000\""
]

PJSUA_CLI_EXPECTS = [
    # Verify the full SDP SIPp sends in its INVITE is received by alt_pjsua.
    # Codec check only (no IP/port) for environment independence.
    [0, r"v=0", ""],
    [0, r"s=-", ""],
    [0, r"t=0 0", ""],
    [0, r"a=rtpmap:0 PCMU/8000", ""],
    [0, r"a=rtpmap:96 H263-1998/90000", ""],
    [0, const.STATE_CONFIRMED, ""],
]
